use crate::vm::mem::DmAllocObject;
use std::collections::HashMap;

#[derive(Debug, PartialEq, Eq)]
pub enum DmType<'a> {
    Byte,
    Int,
    Long,
    Double,
    Boolean,
    ObjectPointer(&'a dyn DmObjectType),
    ByteArray(usize),
    IntArray(usize),
    LongArray(usize),
    DoubleArray(usize),
    BooleanArray(usize),
    ObjectPointerArray(usize, &'a dyn DmObjectType),
    Unit,
}

impl DmType<'_> {
    pub fn size_in_bytes(&self) -> usize {
        match self {
            DmType::Byte => 1,
            DmType::Int => 4,
            DmType::Long => 8,
            DmType::Double => 8,
            DmType::Boolean => 4,
            DmType::ObjectPointer(_) => 8,
            DmType::ByteArray(length) => *length,
            DmType::IntArray(length) => *length * 4,
            DmType::LongArray(length) => *length * 8,
            DmType::DoubleArray(length) => *length * 8,
            DmType::BooleanArray(length) => *length * 4,
            DmType::ObjectPointerArray(length, _) => length * 8,
            DmType::Unit => 0
        }
    }
}

pub trait DmObjectType {
    fn get_fqn(&self) -> &String;
    fn get_method(&self, name: String, self_object: &DmAllocObject) -> Option<&DmFn>;
    fn get_property(&self, name: String, self_object: &DmAllocObject) -> Option<&DmProperty>;
}

#[derive(Debug, Eq)]
pub struct DmFn {
    pub fqn: String,
    pub short_name: String,
    pub byte_code: Vec<u8>,
    pub num_registers: usize,
}

impl PartialEq for DmFn {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

#[derive(Debug, Eq)]
pub struct DmInterface<'a> {
    fqn: String,
    short_name: String,
    properties: HashMap<String, DmProperty<'a>>,
    virtual_methods: HashMap<String, DmVirtualMethod>,
    impl_methods: HashMap<String, DmFn>,
}

impl PartialEq for DmInterface<'_> {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

impl DmObjectType for DmInterface<'_> {
    fn get_fqn(&self) -> &String {
        &self.fqn
    }

    fn get_method(&self, name: String, self_object: &DmAllocObject) -> Option<&DmFn> {
        if self.impl_methods.contains_key(&name) {
            self.impl_methods.get(&name)
        } else if self.virtual_methods.contains_key(&name) {
            self_object.object_type.get_method(name, &self_object)
        } else {
            None
        }
    }

    fn get_property(&self, name: String, self_object: &DmAllocObject) -> Option<&DmProperty> {
        self.properties.get(&name)
    }
}

#[derive(Debug, Eq)]
pub struct DmVirtualMethod {
    pub fqn: String,
    pub short_name: String,
}

impl PartialEq for DmVirtualMethod {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

#[derive(Debug, Eq)]
pub struct DmImplementation<'a> {
    pub fqn: String,
    pub short_name: String,
    pub interface: &'a DmInterface<'a>,
    pub properties: HashMap<String, DmProperty<'a>>,
    pub fields: HashMap<String, DmField<'a>>,
    pub size_in_bytes: usize,
    pub methods: HashMap<String, DmFn>
}

impl PartialEq for DmImplementation<'_> {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

impl DmObjectType for DmImplementation<'_> {
    fn get_fqn(&self) -> &String {
        &self.fqn
    }

    fn get_method(&self, name: String, self_object: &DmAllocObject) -> Option<&DmFn> {
        self.methods.get(&name)
    }

    fn get_property(&self, name: String, self_object: &DmAllocObject) -> Option<&DmProperty> {
        self.properties.get(&name)
    }
}

#[derive(Debug, Eq)]
pub struct DmProperty<'a> {
    pub name: String,
    pub data_offset: usize,
    pub dm_type: DmType<'a>
}

impl PartialEq for DmProperty<'_> {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name
    }
}

#[derive(Debug, Eq)]
pub struct DmField<'a> {
    name: String,
    dm_type: DmType<'a>,
}

impl PartialEq for DmField<'_> {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name
    }
}